@@ -3,24 +3,18 @@ import array
 import operator
 from base64 import b64decode
 from reportlab.pdfgen.canvas import FILL_EVEN_ODD
+import sys
 
 import qrcode
 from reportlab.lib.units import toLength
 
 
 DEFAULT_PARAMS = {
-	'size': '5cm',
 	'padding': '2.5',
-	'fg': '#000000',
-	'bg': None,
 	'version': None,
 	'error_correction': 'L',
-	'x': 0,
-	'y': 0,
-	'invert': False
 }
 FALSE_VALUES = {'off', 'false', 'False', '0', False, 0, None}
-GENERATOR_PARAMS = {'size', 'padding', 'fg', 'bg', 'x', 'y', 'invert'}
 QR_PARAMS = {'version', 'error_correction'}
 QR_ERROR_CORRECTIONS = {
 	'L': qrcode.ERROR_CORRECT_L,
@@ -34,12 +28,11 @@ DIRECTION = (
 	(-1,  0), # left
 	( 0, -1), # up
 )
-# left, direct, right
 DIRECTION_TURNS_CHECKS = (
-	(( 0, -1), ( 0,  0), (-1,  0)), # right
-	(( 0,  0), (-1,  0), (-1, -1)), # down
-	((-1,  0), (-1, -1), ( 0, -1)), # left
-	((-1, -1), ( 0, -1), ( 0,  0)), # up
+	( 0,  0), # right
+	(-1,  0), # down
+	(-1, -1), # left
+	( 0, -1), # up
 )
 
 
@@ -48,21 +41,49 @@ class Vector(tuple):
 		return self.__class__(map(operator.add, self, other))
 
 
+class Transforms:
+	@staticmethod
+	def to_length(val):
+		return toLength(val) if isinstance(val, str) else float(val)
+
+	@staticmethod
+	def to_bool(val):
+		return val not in FALSE_VALUES
+
+	@staticmethod
+	def to_float(val):
+		return float(val)
+
+
 class ReportlabImageBase(qrcode.image.base.BaseImage):
-	size = None
-	padding = None
+	PARAMS = {
+		'size': Transforms.to_length,
+		'padding': None,
+		'fg': None,
+		'bg': None,
+		'x': Transforms.to_length,
+		'y': Transforms.to_length,
+		'invert': Transforms.to_bool,
+		'mask': Transforms.to_bool,
+		'radius': Transforms.to_float,
+		'enhanced_path': Transforms.to_bool,
+	}
+
+	size = toLength('5cm')
+	padding = '2.5'
 	bg = None
-	fg = None
+	fg = '#000000'
 	bitmap = None
 	x = 0
 	y = 0
 	invert = False
+	mask = False
+	enhanced_path = None
+	radius = 0
 
 	def __init__(self, *args, **kwargs):
 		super().__init__(*args, **kwargs)
-		self.invert = self.invert not in FALSE_VALUES
 		self.bitmap = array.array('B', [1 if self.invert else 0] * self.width * self.width)
-		self.size = toLength(self.size) if isinstance(self.size, str) else float(self.size)
 		if isinstance(self.padding, str) and '%' in self.padding:
 			self.padding = float(self.padding[:-1]) * self.size / 100
 		else:
@@ -71,42 +92,106 @@ class ReportlabImageBase(qrcode.image.base.BaseImage):
 				self.padding = (self.size / (self.width + self.padding * 2)) * self.padding
 			except ValueError:
 				self.padding = toLength(self.padding) if isinstance(self.padding, str) else float(self.padding)
-		self.x = toLength(self.x) if isinstance(self.x, str) else float(self.x)
-		self.y = toLength(self.y) if isinstance(self.y, str) else float(self.y)
+		if self.enhanced_path is None:
+			self.enhanced_path = self.radius == 0
 
 	def drawrect(self, row, col):
 		self.bitmap_set((col, row), 0 if self.invert else 1)
 
 	def save(self, stream, kind=None):
-		stream.saveState()
+		if not self.mask:
+			stream.saveState()
+
+		a0, b0, c0, d0, e0, f0 = (1.0, 0.0, 0.0, 1.0, 0.0, 0.0) # current matrix
+
+		def transform(a, b, c, d, e, f):
+			nonlocal a0, b0, c0, d0, e0, f0
+			matrix = (a0*a+c0*b, b0*a+d0*b, a0*c+c0*d, b0*c+d0*d, a0*e+c0*f+e0, b0*e+d0*f+f0)
+			a0, b0, c0, d0, e0, f0 = matrix
+			stream.transform(a, b, c, d, e, f)
+
+		def restore_transform():
+			nonlocal a0, b0, c0, d0, e0, f0
+			det = a0*d0 - c0*b0
+			a = d0/det
+			c = -c0/det
+			e = (c0*f0 - d0*e0)/det
+			d = a0/det
+			b = -b0/det
+			f = (e0*b0 - f0*a0)/det
+			stream.transform(a, b, c, d, e, f)
 
 		try:
 			# Move to start
-			stream.translate(self.x, self.y)
+			transform(1.0, 0.0, 0.0, -1.0, self.x, self.y + self.size)
 
-			# Draw background
-			if self.bg is not None:
-				stream.setFillColor(self.bg)
-				stream.rect(0, 0, self.size, self.size, fill=1, stroke=0)
-
-			# Set foreground
-			stream.setFillColor(self.fg)
+			if not self.mask:
+				self.draw_background(stream)
+				# Set foreground
+				stream.setFillColor(self.fg)
 
 			# Set transform matrix
-			stream.translate(self.padding, self.padding)
-			scale = (self.size - (self.padding * 2)) / self.width
-			stream.scale(scale, scale)
+			scale = (self.size - (self.padding * 2.0)) / self.width
+			transform(scale, 0.0, 0.0, scale, self.padding, self.padding)
 
-			# Draw code
 			p = stream.beginPath()
-			for segment in self.get_segments():
-				p.moveTo(segment[0][0], self.width - segment[0][1])
-				for coords in segment[1:-1]:
-					p.lineTo(coords[0], self.width - coords[1])
-				p.close()
-			stream.drawPath(p, stroke=0, fill=1, fillMode=FILL_EVEN_ODD)
+			if self.radius == 0:
+				p = self.draw_code(p)
+			else:
+				p = self.draw_rounded_code(p)
+			if self.mask:
+				stream.clipPath(p, stroke=0)
+			else:
+				stream.drawPath(p, stroke=0, fill=1, fillMode=FILL_EVEN_ODD)
 		finally:
-			stream.restoreState()
+			if self.mask:
+				restore_transform()
+			else:
+				stream.restoreState()
+
+	def draw_background(self, stream):
+		"""
+		Draw rectangle on background if is not transparent
+		"""
+		if self.bg is not None:
+			stream.setFillColor(self.bg)
+			stream.rect(0, 0, self.size, self.size, fill=1, stroke=0)
+
+	def draw_code(self, p):
+		"""
+		Draw QR code
+		"""
+		for segment in self.get_segments():
+			p.moveTo(segment[0][0], segment[0][1])
+			for coords in segment[1:-1]:
+				p.lineTo(coords[0], coords[1])
+			p.close()
+		return p
+
+	def draw_rounded_code(self, p):
+		"""
+		Draw QR code using rounded paths
+		"""
+		for segment in self.get_segments():
+			segment = segment[:-1]
+			for i in range(0, len(segment)):
+				coords = segment[i]
+				prev_coords = segment[i - 1]
+				next_coords = segment[(i + 1) % len(segment)]
+				prev_dir = self.__calc_round_direction(prev_coords, coords, self.radius)
+				next_dir = self.__calc_round_direction(next_coords, coords, self.radius)
+				if i == 0:
+					p.moveTo(coords[0] + prev_dir[0], coords[1] + prev_dir[1])
+				else:
+					p.lineTo(coords[0] + prev_dir[0], coords[1] + prev_dir[1])
+				c = 0.45 # 1 - (4/3)*tan(pi/8)
+				p.curveTo(
+					coords[0] + prev_dir[0] * c, coords[1] + prev_dir[1] * c,
+					coords[0] + next_dir[0] * c, coords[1] + next_dir[1] * c,
+					coords[0] + next_dir[0], coords[1] + next_dir[1],
+				)
+			p.close()
+		return p
 
 	def addr(self, coords):
 		"""
@@ -169,13 +254,6 @@ class ReportlabImageBase(qrcode.image.base.BaseImage):
 			# Or no pixels left
 			return
 
-		# Accumulated path
-		path = []
-		# Begin of line
-		path.append(tuple(coords))
-		# Default direction to right
-		direction = 0
-
 		def move():
 			nonlocal coords
 			step = DIRECTION[direction]
@@ -190,28 +268,43 @@ class ReportlabImageBase(qrcode.image.base.BaseImage):
 			# Step
 			coords += step
 
+		# Accumulated path
+		path = []
+		# Begin of line
+		path.append(tuple(coords))
+		# Default direction to right
+		direction = 0
+		# Default clockwiese direction
+		clockwiese = 1
+
 		# Move to right
 		move()
 
 		# From shape begin to end
 		while coords != path[0]:
 			# Trun left
-			val = self.bitmap_get(coords + DIRECTION_TURNS_CHECKS[direction][0])
+			val = self.bitmap_get(coords + DIRECTION_TURNS_CHECKS[(direction - max(0, clockwiese)) % 4])
 			if val:
+				if self.enhanced_path:
+					# Detect intersection pattern and change direction
+					if not self.bitmap_get(coords + DIRECTION_TURNS_CHECKS[(direction + min(0, clockwiese)) % 4]):
+						move()
+						clockwiese = -clockwiese
+						continue;
 				path.append(tuple(coords))
-				direction = (direction - 1) % 4
+				direction = (direction - clockwiese) % 4
 				move()
 				continue
 
 			# Straight
-			val = self.bitmap_get(coords + DIRECTION_TURNS_CHECKS[direction][1])
+			val = self.bitmap_get(coords + DIRECTION_TURNS_CHECKS[(direction + min(0, clockwiese)) % 4])
 			if val:
 				move()
 				continue
 
 			# Trun right
 			path.append(tuple(coords))
-			direction = (direction + 1) % 4
+			direction = (direction + clockwiese) % 4
 			move()
 
 		path.append(tuple(coords))
@@ -225,13 +318,27 @@ class ReportlabImageBase(qrcode.image.base.BaseImage):
 
 		return path
 
+	def __calc_round_direction(self, src, dst, radius):
+		return [min(max((s - d) * 0.5, -radius), radius) for s, d in zip(src, dst)]
 
 
-def reportlab_image_factory(**kwargs):
+def reportlab_image_factory(base=ReportlabImageBase, **kwargs):
 	"""
 	Returns ReportlabImage class for qrcode image_factory
 	"""
-	return type('ReportlabImage', (ReportlabImageBase,), kwargs)
+	params = {}
+	# Chceck each parameter
+	for key, value in kwargs.items():
+		# If is unknown, raise exception
+		if key not in base.PARAMS:
+			raise ValueError("Unknown attribute '%s'" % key)
+		# Get transform function like to int
+		transform = base.PARAMS[key]
+		try:
+			params[key] = value if transform is None else transform(value)
+		except Exception as e:
+			raise ValueError("Wrong value '%s' for attribute %s: %s" % (value, key, e))
+	return type('ReportlabImage', (base,), params)
 
 
 def clean_params(params):
@@ -239,10 +346,6 @@ def clean_params(params):
 	Validate and clean parameters
 	"""
 
-	for key, __ in params.items():
-		if key not in GENERATOR_PARAMS and key not in QR_PARAMS:
-			raise ValueError("Unknown attribute '%s'" % key)
-
 	if params['version'] is not None:
 		try:
 			params['version'] = int(params['version'])
@@ -255,7 +358,7 @@ def clean_params(params):
 		raise ValueError("Unknown error correction '%s', expected one of %s" % (params['error_correction'], ', '.join(QR_ERROR_CORRECTIONS.keys())))
 
 
-def parse_graphic_params(params):
+def parse_params_string(params):
 	"""
 	Parses params string in form:
 
@@ -292,14 +395,11 @@ def parse_graphic_params(params):
 
 
 def build_qrcode(params, text):
-	factory_kwargs = {key: value for key, value in params.items() if key in GENERATOR_PARAMS}
+	factory_kwargs = {key: value for key, value in params.items() if key not in QR_PARAMS}
 	qr_kwargs = {key: value for key, value in params.items() if key in QR_PARAMS}
-	return qrcode.make(text, image_factory=reportlab_image_factory(**factory_kwargs), border=0, **qr_kwargs)
-
-
-def qr_factory(params):
-	params, text = parse_graphic_params(params)
-	return build_qrcode(params, text)
+	qr = qrcode.QRCode(image_factory=reportlab_image_factory(**factory_kwargs), border=0, **qr_kwargs)
+	qr.add_data(text)
+	return qr.make_image()
 
 
 def qr_draw(canvas, text, **kwargs):
@@ -321,4 +421,4 @@ def qr(canvas, params=None):
 		<plugInGraphic module="reportlab_qrcode" function="qr">size=5cm;text;Simple text</plugInGraphic>
 	</illustration>
 	"""
-	qr_factory(params).save(canvas)
+	build_qrcode(*parse_params_string(params)).save(canvas)
